@@ -49,6 +49,8 @@
 #include <QTextDocument> // for Qt::escape
 #include <QLocale>
 #include <QIcon>
+#include <QColor>
+#include <QRegularExpression>
 
 #include <gpgme++/gpgmepp_version.h>
 #if GPGMEPP_VERSION > 0x10900
@@ -257,6 +259,19 @@ QString format_keytype(const Key &key)
     }
 }
 
+QString format_subkeytype(const Subkey &subkey)
+{
+    const auto algo = subkey.publicKeyAlgorithm();
+
+    if (algo == Subkey::AlgoECC ||
+        algo == Subkey::AlgoECDSA ||
+        algo == Subkey::AlgoECDH ||
+        algo == Subkey::AlgoEDDSA) {
+        return QString::fromStdString(subkey.algoName());
+    }
+    return i18n("%1-bit %2", subkey.length(), QLatin1String(subkey.publicKeyAlgorithmAsString()));
+}
+
 QString format_keyusage(const Key &key)
 {
     QStringList capabilities;
@@ -279,6 +294,28 @@ QString format_keyusage(const Key &key)
     return capabilities.join(QStringLiteral(", "));
 }
 
+QString format_subkeyusage(const Subkey &subkey)
+{
+    QStringList capabilities;
+    if (subkey.canSign()) {
+        if (subkey.isQualified()) {
+            capabilities.push_back(i18n("Signing (Qualified)"));
+        } else {
+            capabilities.push_back(i18n("Signing"));
+        }
+    }
+    if (subkey.canEncrypt()) {
+        capabilities.push_back(i18n("Encryption"));
+    }
+    if (subkey.canCertify()) {
+        capabilities.push_back(i18n("Certifying User-IDs"));
+    }
+    if (subkey.canAuthenticate()) {
+        capabilities.push_back(i18n("SSH Authentication"));
+    }
+    return capabilities.join(QStringLiteral(", "));
+}
+
 static QString time_t2string(time_t t)
 {
     QDateTime dt;
@@ -394,6 +431,40 @@ QString Formatting::toolTip(const Key &key, int flags)
             result += format_row(i18n("Stored"), i18nc("stored...", "on this computer"));
         }
     }
+    if (flags & Subkeys) {
+        for (const auto &sub: key.subkeys()) {
+            result += QLatin1String("<hr/>");
+            result += format_row(i18n("Subkey"), sub.fingerprint());
+            if (sub.isRevoked()) {
+                result += format_row(i18n("Status"), i18n("Revoked"));
+            } else if (sub.isExpired()) {
+                result += format_row(i18n("Status"), i18n("Expired"));
+            }
+            if (flags & ExpiryDates) {
+                result += format_row(i18n("Created"), time_t2string(sub.creationTime()));
+
+                if (key.isExpired()) {
+                    result += format_row(i18n("Expired"), time_t2string(sub.expirationTime()));
+                } else if (!subkey.neverExpires()) {
+                    result += format_row(i18n("Expires"), time_t2string(sub.expirationTime()));
+                }
+            }
+
+            if (flags & CertificateType) {
+                result += format_row(i18n("Type"), format_subkeytype(sub));
+            }
+            if (flags & CertificateUsage) {
+                result += format_row(i18n("Usage"), format_subkeyusage(sub));
+            }
+            if (flags & StorageLocation) {
+                if (const char *card = sub.cardSerialNumber()) {
+                    result += format_row(i18n("Stored"), i18nc("stored...", "on SmartCard with serial no. %1", QString::fromUtf8(card)));
+                } else {
+                    result += format_row(i18n("Stored"), i18nc("stored...", "on this computer"));
+                }
+            }
+        }
+    }
     result += QLatin1String("</table>");
 
     return result;
@@ -587,10 +658,17 @@ QString Formatting::validityShort(const UserID::Signature &sig)
     switch (sig.status()) {
     case UserID::Signature::NoError:
         if (!sig.isInvalid()) {
-            if (sig.certClass() > 0) {
+            /* See RFC 4880 Section 5.2.1 */
+            switch (sig.certClass()) {
+            case 0x10: /* Generic */
+            case 0x11: /* Persona */
+            case 0x12: /* Casual */
+            case 0x13: /* Positive */
+                return i18n("valid");
+            case 0x30:
+                return i18n("revoked");
+            default:
                 return i18n("class %1", sig.certClass());
-            } else {
-                return i18nc("good/valid signature", "good");
             }
         }
         Q_FALLTHROUGH();
@@ -600,11 +678,43 @@ QString Formatting::validityShort(const UserID::Signature &sig)
     case UserID::Signature::SigExpired:   return i18n("expired");
     case UserID::Signature::KeyExpired:   return i18n("certificate expired");
     case UserID::Signature::BadSignature: return i18nc("fake/invalid signature", "bad");
-    case UserID::Signature::NoPublicKey:  return QString();
+    case UserID::Signature::NoPublicKey:  return i18n("no public key");
     }
     return QString();
 }
 
+QIcon Formatting::validityIcon(const UserID::Signature &sig)
+{
+    switch (sig.status()) {
+    case UserID::Signature::NoError:
+        if (!sig.isInvalid()) {
+            /* See RFC 4880 Section 5.2.1 */
+            switch (sig.certClass()) {
+            case 0x10: /* Generic */
+            case 0x11: /* Persona */
+            case 0x12: /* Casual */
+            case 0x13: /* Positive */
+                return QIcon::fromTheme(QStringLiteral("emblem-success"));
+            case 0x30:
+                return QIcon::fromTheme(QStringLiteral("emblem-error"));
+            default:
+                return QIcon();
+            }
+        }
+        Q_FALLTHROUGH();
+        // fall through:
+    case UserID::Signature::BadSignature:
+    case UserID::Signature::GeneralError:
+        return QIcon::fromTheme(QStringLiteral("emblem-error"));
+    case UserID::Signature::SigExpired:
+    case UserID::Signature::KeyExpired:
+        return QIcon::fromTheme(QStringLiteral("emblem-information"));
+    case UserID::Signature::NoPublicKey:
+        return QIcon::fromTheme(QStringLiteral("emblem-question"));
+    }
+    return QIcon();
+}
+
 QString Formatting::formatKeyLink(const Key &key)
 {
     if (key.isNull()) {
@@ -893,3 +1003,11 @@ QString Formatting::complianceStringShort(const GpgME::Key &key)
 
     return i18nc("As in not all user IDs are valid.", "not certified");
 }
+
+QString Formatting::prettyID(const char *id)
+{
+    if (!id) {
+        return QString();
+    }
+    return QString::fromLatin1(id).toUpper().replace(QRegularExpression("(....)"), "\\1 ").trimmed();
+}
